#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-

## Copyright 2004 by LivingLogic AG, Bayreuth/Germany.
## Copyright 2004 by Walter Drwald
##
## All Rights Reserved
##
## Permission to use, copy, modify, and distribute this software and its documentation
## for any purpose and without fee is hereby granted, provided that the above copyright
## notice appears in all copies and that both that copyright notice and this permission
## notice appear in supporting documentation, and that the name of LivingLogic AG or
## the author not be used in advertising or publicity pertaining to distribution of the
## software without specific, written prior permission.
##
## LIVINGLOGIC AG AND THE AUTHOR DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
## INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
## LIVINGLOGIC AG OR THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
## DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
## IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
## IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.


import datetime


import cx_Oracle


class Proc:
	_ora2cx = {
		"number": cx_Oracle.NUMBER,
		"varchar2": cx_Oracle.STRING
	}
	
	def __init__(self, name):
		self.name = name
		self._realargs = None

	def __repr__(self):
		return "Proc(%r)" % self.name

	def _calcrealargs(self, cursor):
		if self._realargs is None:
			self._realargs = ({}, [])
			cursor.execute("select lower(argument_name), lower(in_out), lower(data_type) from user_arguments where lower(object_name)=lower(:name) order by sequence", name=self.name)
			for (i, (name, in_out, datatype)) in enumerate(cursor):
				self._realargs[1].append((name, datatype, "in" in in_out, "out" in in_out))
				self._realargs[0][name] = (i, datatype, "in" in in_out, "out" in in_out)

	def __call__(self, cursor, **kwargs):
		self._calcrealargs(cursor)
		realargs = [None]*len(self._realargs[1])
		for (key, value) in kwargs.iteritems():
			(pos, datatype, isin, isout) = self._realargs[0][key.lower()]
			if isinstance(value, datetime.datetime):
				value = cx_Oracle.Timestamp(value.year, value.month, value.day, value.hour, value.minute, value.second)
			elif isinstance(value, datetime.date):
				value = cx_Oracle.Date(value.year, value.month, value.day)
			elif isinstance(value, unicode):
				value = value.encode("iso-8859-1")
			elif isinstance(value, long):
				value = int(value)
				if isinstance(value, long):
					value = float(value)
			realargs[pos] = value

		# Replace out parameters with variables
		for (pos, (name, datatype, isin, isout)) in enumerate(self._realargs[1]):
			if isout:
				var = cursor.var(self._ora2cx[datatype])
				var.setvalue(0, realargs[pos])
				realargs[pos] = var

		result = Record()
		for (i, value) in enumerate(cursor.callproc(self.name, realargs)):
			if isinstance(value, cx_Oracle.TIMESTAMP):
				value = datetime.datetime(value.year, value.month, value.day, value.hour, value.minute, value.second, value.fsecond)
			elif isinstance(value, cx_Oracle.DATETIME):
				value = datetime.datetime(value.year, value.month, value.day, value.hour, value.minute, value.second)
			result[self._realargs[1][i][0]] = value
		return result


class LLProc(Proc):
	def __call__(self, cursor, **kwargs):
		args = {}
		for (key, value) in kwargs.iteritems():
			if key == "user":
				key = "c_user"
			else:
				key = "p_%s" % key
			args[key] = value

		result = Record()
		for (key, value) in Proc.__call__(self, cursor, **args).iteritems():
			if key == "c_user":
				key = "user"
			elif key.startswith("p_"):
				key = key[2:]
			else:
				raise ValueError("unknown parameter name %r in result" % key)
			result[key] = value
		return result


class Record(dict):
	def makerecord(cls, cursor, row):
		record = cls()
		for (i, field) in enumerate(row):
			field = row[i]
			if isinstance(field, cx_Oracle.TIMESTAMP):
				field = datetime.datetime(field.year, field.month, field.day, field.hour, field.minute, field.second, field.fsecond)
			elif isinstance(field, cx_Oracle.DATETIME):
				field = datetime.datetime(field.year, field.month, field.day, field.hour, field.minute, field.second)
			record[cursor.description[i][0]] = field
		return record
	makerecord = classmethod(makerecord)

	def fetchone(cls, cursor):
		record = cursor.fetchone()
		if record is not None:
			record = cls.makerecord(cursor, record)
		return record
	fetchone = classmethod(fetchone)

	def fetchmany(cls, cursor, rows=None):
		if rows is None:
			return [ cls.makerecord(cursor, row) for row in cursor.fetchmany()]
		else:
			return [ cls.makerecord(cursor, row) for row in cursor.fetchmany(rows)]
	fetchmany = classmethod(fetchmany)

	def fetchall(cls, cursor):
		return [ cls.makerecord(cursor, row) for row in cursor.fetchall()]
	fetchall = classmethod(fetchall)

	def fetch(cls, cursor):
		for row in cursor:
			yield cls.makerecord(cursor, row)
	fetch = classmethod(fetch)

	def __getitem__(self, name):
		return dict.__getitem__(self, name.lower())

	def __setitem__(self, name, value):
		dict.__setitem__(self, name.lower(), value)

	def __delitem__(self, name):
		dict.__delitem__(self, name.lower())

	def __getattr__(self, name):
		return self.__getitem__(name)

	def __setattr__(self, name, value):
		self.__setitem__(name, value)

	def __delattr__(self, name):
		self.__delitem__(name)

	def __repr__(self):
		return "%s.%s(%s)" % (self.__class__.__module__, self.__class__.__name__, ", ".join(["%s=%r" % item for item in self.iteritems()]))
